"""
Product location service.

Given a product locator, the location service produces the list of files and their locations in NGAS.
The primary user of this information is the archive data fetcher.
"""
from functools import reduce
from pathlib import Path

from pyat.schema import ScienceProduct, AncillaryProduct
from pyat.schema.ngasmodel import NGASFile, NGASDisk
from pyramid.httpexceptions import exception_response
from pyramid.view import view_config
from sqlalchemy import func

from services import DB


@view_config(route_name='product_location_service', renderer='json')
def product_location_service(request):
    # obtain the product locator from the request and pass it along
    return product_locations(request.params.get('locator'))


def product_locations(locator):
    """
    Given a product locator, return a dictionary with the files
    and metadata the data fetcher needs.
    """

    # does this look like an ancillary product or not?
    product = lookup_product_by_locator(locator)

    # ensure we got a product
    if product is None:
        raise exception_response(404, detail=f"Product with this locator not found")

    # start building the outer part of the result:
    files = product.filegroup.all_files

    # group the files by what cluster they belong to
    file_clusters = partition(lambda f: f.ngas_cluster, files)

    # for each cluster, obtain the "result-flavored" dictionaries for the
    # files which belong to that cluster
    result_files = []
    for cluster, files in file_clusters.items():
        result_files.extend(ngas_metadata_for(files, cluster, product))

    # produce the final result by adding up the total sizes of each file
    return {'files': result_files,
            'aggregate_size': sum(f['size'] for f in result_files)}


def lookup_product_by_locator(locator):
    """
    Returns an Ancillary Product or a Science Product, as appropriate, for this locator.
    :param locator:  the locator to look up
    :return: a Product of some kind
    """

    # This kind of locator inspection is *only* OK inside this service!
    if 'ancillary_product' in locator:
        # obtain the ancillary product
        product = DB['SDM'].query(AncillaryProduct).get(locator)
    else:
        # obtain the science product
        product = DB['SDM'].query(ScienceProduct).get(locator)
    return product


def ngas_metadata_for(files, cluster, product):
    """
    Retrieves the "result-shaped" dictionaries for these files by querying NGAS for the details.

    :param files:    the files to process
    :param cluster:  the cluster they belong to
    :param product:  the product this is all happening under
    :return:         a list of "result-flavored" dictionaries
    """
    NGAS = get_ngas(cluster)

    # need this map to hook up NGAS database NGASFiles with archive database Files
    ngasmap = dict((f.ngas_id, f) for f in files)

    max_versions = NGAS.query(NGASFile.file_id, func.max(NGASFile.version).label('latest')) \
        .filter(NGASFile.file_id.in_(ngasmap.keys())).group_by(NGASFile.file_id).subquery()
    query = NGAS.query(NGASFile, max_versions, NGASDisk) \
        .filter(NGASFile.file_id.in_(ngasmap.keys()),
                max_versions.c.file_id == NGASFile.file_id,
                max_versions.c.latest == NGASFile.version,
                NGASDisk.disk_id == NGASFile.disk_id)
    nfiles = [r[0] for r in query]
    # make sure we have exactly the same number of files as we expect
    if len(nfiles) < len(files):
        missing = len(files) - len(nfiles)
        raise exception_response(500,
                                 detail=f"Missing NGAS locations for {missing} files")

    # OK, we do, so let's map everything together
    results = []
    for nfile in nfiles:
        # "join" up the file
        file = ngasmap[nfile.file_id]

        # get the server
        nserver = nfile.disk.host

        # prepare the result for this file
        server = dict(server=f'{nserver.server}:{nserver.port}',
                      location=file.ngas_location,
                      cluster=file.ngas_cluster)

        if nfile.format == 'application/x-bdf':
            relative_path = 'ASDMBinary/' + Path(nfile.file_id).stem
        else:
            relative_path = file.filename

        rfile = dict(ngas_file_id=nfile.file_id,
                     subdirectory=product.subdirectory,
                     relative_path=relative_path,
                     checksum=nfile.checksum,
                     checksum_type=nfile.checksum_plugin,
                     version=nfile.version,
                     size=nfile.file_size,
                     server=server)
        results.append(rfile)
    return results


def partition(f, items):
    """
    Partition items into separate groups for each kind of result you get
    from applying f to an item. For instance:

    >>> def even(x): return 'even' if x % 2 == 0 else 'odd'
    >>> partition(even, range(1,5))
    {'odd': [1, 3], 'even': [2, 4]}

    :param f:     the partitioning function
    :param items: the items to partition
    :return:      a dictionary of partitions
    """
    result = {}
    for item in items:
        result.setdefault(f(item), []).append(item)
    return result


def get_ngas(cluster):
    """
    Returns the appropriate database connection for the cluster argument.

    :param cluster:  the cluster to identify
    :return:         a SqlAlchemy session for this cluster
    """
    if cluster == 'DSOC':
        return DB['NGAS']
    else:
        # A small hack here, to ensure we have the right "schema"
        DB['NGAS_NAASC'].execute('ALTER SESSION SET CURRENT_SCHEMA = NGAS')
        return DB['NGAS_NAASC']
